--[[
Copyright (c) 2015 - 2016 Robert Machmer

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the
Software without restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]
local dir	= 'Graphoon'
local Node	= require( dir .. '.Node' )
local gr, floor, random = love.graphics, math.floor, love.math.random
local pi, sin, cos, min, max = math.pi, math.sin, math.cos, math.min, math.max
gr.setLineWidth(2)
local load = require( 'load' )	-- loads selected data

local Graph = {}
Graph.__index = Graph

function Graph.new(  dir )
	local tab = { nodes = {}, edges = {}, edgeIDs = 0 }
	setmetatable( tab, Graph )
	local edg, nod = load.data( dir..'/edges.txt', 'edges' )	-- loads node interconnections
	calls = load.data( dir..'/calls.txt', 'calls' )				-- [ loads function calls ]

	for i = 1, #nod do		-- create nodes
        tab.nodes[nod[i]] = tab:addNode( nod[i],random( 0, gr.getWidth() ), random( 0, gr.getHeight() ))
		end

	for i = 1, #edg do		-- create edges
		tab:addEdge( tab.nodes[edg[i].origin], tab.nodes[edg[i].target])
		end

    tab:setAnchor( nod[1], gr.getWidth() * 0.5, gr.getHeight() * 0.5 )
	tab.nodes[nod[1]].color = color.center	-- color and anchor first node
	return tab
	end

function Graph.updateBoundaries( nx, ny )	-- update area enclosing graph
        return min( minX or nx, nx ), max( maxX or nx, nx ), min( minY or ny, ny ), max( maxY or ny, ny )
		end

function Graph:addEdge( origin, target )	-- create edge between two nodes
		for i = 1, #self.edges do			-- don't create new edge if already connected
            if self.edges.origin == origin and self.edges.target == target then return end
			end

		if origin ~= target then
			self.edges[self.edgeIDs] = { id = id, origin = origin, target = target }
			self.edgeIDs = self.edgeIDs + 1	-- note: edges are numbered, not IDed
			end
		end

    -- Adds a node to the graph.
    -- @param id - The ID will be used to reference the Node inside of the graph.
    -- @param x  - The x coordinate the Node should be spawned at (optional).
    -- @param y  - The y coordinate the Node should be spawned at (optional).
    -- @param anchor - Wether the node should be locked in place or not (optional).

	function Graph:addNode( id, x, y, anchor, color, size, ... )
        self.nodes[id] = Node.new( id, x, y, anchor, color, size, ... )
        return self.nodes[id]
		end

    function Graph:removeNode( node )		-- remove node and attached edges
        if node.od then		-- node name changed but node.id is the original id
			self.nodes[node.od] = nil
			else
			self.nodes[node.id] = nil
			end
		end

	function Graph:removeEdges( node )	-- remove edges attached to a deleted node
        for id, edge in pairs( self.edges ) do
            if edge.origin.id == node.id or edge.target.id == node.id then
                self.edges[id] = nil
				end
			end
		end

	function Graph:replaceEdgeName( old, new )	-- renames nodes used in edges
		for i = 1, #self.edges do
			if self.edges[i].origin.id == old then self.edges[i].origin.id = new end
			if self.edges[i].target.id == old then self.edges[i].target.id = new end
			end
		end

    function Graph:update( dt )	-- update graph
        for _, edge in pairs( self.edges ) do
            edge.origin:attractTo( edge.target )
            edge.target:attractTo( edge.origin )
			end

		minX, maxX, minY, maxY = nil, nil, nil, nil

        for _, nodeA in pairs( self.nodes ) do
            if not nodeA.anchor then
                for _, nodeB in pairs( self.nodes ) do
                    if nodeA ~= nodeB then
                        nodeA:repelFrom( nodeB )
						end
					end
                nodeA:move( dt )
				end

            minX, maxX, minY, maxY = Graph.updateBoundaries( nodeA.x,nodeA.y )
			end
		end

	 function Graph:draw(  )		-- drawing functions
		if active.e then
			for _, edge in pairs( self.edges ) do
				gr.setColor( color.edge )
				gr.line( edge.origin.x, edge.origin.y, edge.target.x, edge.target.y )
				gr.line( edge.origin.x-2, edge.origin.y-2, edge.target.x, edge.target.y )
				gr.line( edge.origin.x+2, edge.origin.y+2, edge.target.x, edge.target.y )
				end
			end
		if calls then
			for i = 1, #calls do
				if self.nodes[calls[i][1]] and  self.nodes[calls[i][2]] then
					local call = {
						self.nodes[calls[i][1]].x, self.nodes[calls[i][1]].y,
						self.nodes[calls[i][2]].x, self.nodes[calls[i][2]].y}
					if active.c and call then
						gr.setColor( color.calls )
						gr.line( call[1],call[2],call[3],call[4])
						end
					end
				end
			end

		for _, node in pairs( self.nodes ) do
			local x, y = floor( node.x ), floor( node.y )
			if node.color then gr.setColor(node.color) else gr.setColor( color.node ) end
			if active.p then
				gr.draw( sprite.image, x, y, 0, node.size/sprite.ht, node.size/sprite.ht, sprite.offset, sprite.offset )
				end
			if node.anchor then
				gr.draw(anchor.image,node.x,node.y,-rotate.angle-pi/4,anchor.scale,anchor.scale,anchor.w/2,anchor.h/2)
				end
			if active.l then
				gr.setColor( color.text )
				gr.print( ' '..node.id, x , y , -rotate.angle ,1,1,-fontsize/2,fontsize/2)
				end
			end
		end

    function Graph:getNode( id ) return self.nodes[id] end

	function Graph:getNodeAt( x, y, range )	-- finds closest node to cursor
        for _, node in pairs( self.nodes ) do
            local nx, ny = node.x, node.y
            if x < nx + range and x > nx - range and y < ny + range and y > ny - range then
                return node
				end
			end
		end

    function Graph.getBoundaries() return minX, maxX, minY, maxY end

    function Graph:setAnchor( id, x, y )	-- node stationary, unmoveable by applied force
		self.nodes[id].x, self.nodes[id].y, self.nodes[id].anchor = x, y, true
		end

	function Graph.showBoundingBox()
		gr.setColor(color.bounding)
		local mX, mY = minX - sprite.size/2, minY - sprite.size/2	-- keep nodes in box
		local wX, hY = maxX - minX + sprite.size, maxY - minY + sprite.size
		gr.rectangle( 'line', mX, mY, wX, hY )
		gr.circle( 'fill', wX/2+mX, hY/2+mY, 5 )	-- center of box
		end

	function Graph.showStats( x, y, s )
		local stats = gr.getStats()
		x = x or 20
		y = y or 500
		s = s or gr.getFont():getHeight()
		gr.setColor(color.controls)
		gr.rectangle('fill',x-10,y-10+s,300,185,10,10)
		gr.setColor( color.text )
		for i, j in pairs( stats ) do
			y = y + s
			gr.print( ( '%s:  %s' ):format( i, j ), x, y )
			end
		gr.print( ('fps: %s' ):format(love.timer.getFPS()), x, y + s)
		gr.print( ('memory use: %s Kbyes' ):format( floor(collectgarbage("count")))..' Kbytes',x,y + 2*s )
		end

	function Graph:showKeylog( x, y )
		local x, y, st = x or 10, y or 30
		gr.setColor(color.controls)
		local sy = #keylog < 20 and #keylog or 20	-- limit display to last 20 entries
		gr.rectangle( 'fill', x-10, y, 450, sy * 20, 10, 10 )
		gr.setColor(color.text)
		if #keylog < 20 then st = 1 else st = #keylog - 19 end
		for i = st, #keylog do gr.print( keylog[i], x, (i - st  ) * 20 + y ) end
		end

	function Graph.save( filename )
		local screenshot = gr.newScreenshot()
		screenshot:encode( 'png', filename .. '.png' )
		end

	function Graph:nodeData( node )
		local x,y,s = 885,550,20
		gr.setColor( color.controls )
		gr.rectangle('fill',875,545,250,150,10,10)
		gr.setColor( color.text )
		gr.print( ( 'Node: %s' ):format( node.id ),x,y )
		gr.print( ( 'X: %s' ):format( floor(node.x) ),x,y+s )
		gr.print( ( 'Y: %s' ):format( floor(node.y) ),x,y+2*s )
		gr.print( ( 'Mass: %s' ):format( node.mass ),x,y+3*s )
		gr.print( ( 'vx: %s' ):format( node.vx),x,y+4*s )
		gr.print( ( 'vy: %s' ):format( node.vy ),x,y+5*s )
		gr.print( ( 'Size: %s' ):format( node.size ),x,y+6*s )
		end

	function Graph.showOrientation(rot,cx,cy)
		gr.setColor( color.rotation )
		local c, s		= cos( rot - pi/4 ), sin( rot - pi/4 )
		local sn		= {25*(c-s), 25*(s+c)}
		local x1,y1,x2,y2 = cx+sn[1], cy+sn[2], cx-sn[1], cy-sn[2]
		local x3,y3,x4,y4 = cx-sn[2], cy+sn[1], cx+sn[2], cy-sn[1]
		gr.circle( 'fill', x4, y4, 5 )
		gr.line( x1,y1,x2,y2 )
		gr.line( x3,y3,x4,y4 )
		end

return Graph
